@@ -4,10 +4,16 @@ import math
 
 fwd_kernels = dict()
 fwd_src = '''
-__global__ void softmax_fwd(TYPE *X, float scale,
-                            int *LUT, TYPE *KP_M, TYPE *ATTN_M,
-                            int num_blocks, int sizemax, 
-                            long stride_zx, int stride_zkpm, int stride_zattnm){ 
+__global__ void softmax_fwd(TYPE *X __readonly __noalias __aligned(16), 
+                            float scale,
+                            int *LUT __readonly __noalias __aligned(16), 
+                            TYPE *KP_M __readonly __noalias __aligned(16), 
+                            TYPE *ATTN_M __readonly __noalias __aligned(16),
+                            int num_blocks, 
+                            int sizemax, 
+                            long stride_zx __multipleof(BLOCK), 
+                            int stride_zkpm __multipleof(BLOCK), 
+                            int stride_zattnm __multipleof(BLOCK)){ 
   int pidhm = get_program_id(0);
   int pidz = get_program_id(1);
 
@@ -26,27 +32,28 @@ __global__ void softmax_fwd(TYPE *X, float scale,
   int   rbmn[TN] = check ? rbn : size - 1;
 
   // block id and column id
-  long blockid [TN]  = *(LUT + offset + rbmn);
-  long columnid[TN]  = *(LUT + offset + rbmn + num_blocks);
-  long rowid   [TN]  = *(LUT + offset + rbmn + num_blocks*2);
-
-  // pointers to key padding mask
-  TYPE* pkp_m[TN]  = KP_M + pidz * stride_zkpm 
-                              + columnid * BLOCK
-                              + rxn;
-
-  // pointers to attention mask
-  TYPE* pattn_m[TN] = ATTN_M + columnid * BLOCK 
-                                 + rowid * BLOCK * stride_zattnm
-                                 + rxm * stride_zattnm
-                                 + rxn;
+  long blockid [TN]  = *(LUT + offset + rbmn*3 + 0);
+  long columnid[TN]  = *(LUT + offset + rbmn*3 + 1);
+  long rowid   [TN]  = *(LUT + offset + rbmn*3 + 2);
 
   // pointers to X
   TYPE* px[TN]  = X + pidz * stride_zx
-                        + blockid * BLOCK * BLOCK 
-                        + rxm * BLOCK 
-                        + rxn;
+                    + blockid * BLOCK * BLOCK 
+                    + rxm * BLOCK 
+                    + rxn;
 
+  // pointers to key padding mask
+  TYPE* pkp_m[TN]  = KP_M + pidz * stride_zkpm 
+                          + columnid * BLOCK
+                          + rxn;
+
+  // pointers to attention mask
+  TYPE* pattn_m[TN] = ATTN_M + columnid * BLOCK 
+                             + rowid * BLOCK * stride_zattnm
+                             + rxm * stride_zattnm
+                             + rxn;
+
+ 
   // load  input
   TYPE x[TN] =  check ? *px : -INFINITY;
   // load key-padding mask
@@ -83,9 +90,13 @@ __global__ void softmax_fwd(TYPE *X, float scale,
 bwd_kernels = dict()
 bwd_src = '''
 
-__global__ void softmax_bwd(TYPE * X, float scale,
-                            TYPE* DX, int* LUT,
-                            int sizemax, long stride_zx, long stride_zdx) {
+__global__ void softmax_bwd(TYPE * X __readonly __noalias __aligned(16), 
+                            float scale,
+                            TYPE* DX __readonly __noalias __aligned(16), 
+                            int* LUT,
+                            int sizemax, 
+                            long stride_zx __multipleof(BLOCK), 
+                            long stride_zdx __multipleof(BLOCK)) {
     int pidhm = get_program_id(0);
     int pidz = get_program_id(1);
 
@@ -105,7 +116,7 @@ __global__ void softmax_bwd(TYPE * X, float scale,
     int rbmn[TN] = check ? rbn : size - 1;
 
     // initialize pointers to block-sparse input
-    long blockid[TN] = *(LUT + offset + rbmn);
+    long blockid[TN] = *(LUT + offset + rbmn*3);
 
     TYPE* px[TN] = X + pidz * stride_zx
                          + blockid * BLOCK * BLOCK
@@ -120,9 +131,12 @@ __global__ void softmax_bwd(TYPE * X, float scale,
     // compute fused softmax backward
     TYPE x[TN] = check ? *px : 0;
     TYPE dx[TN] = check ? *pdx : 0;
-    TYPE xdx[TN] = x * dx;
-    TYPE xdxsum = (check ? xdx : 0)[+];
-    TYPE y[TN] = x * (dx - xdxsum) * scale;
+    float Fdx[TN] = dx;
+    float Fx[TN] = x;
+    float Fxdx[TN] = Fdx*Fx;
+    float Fxdxsum = Fxdx[+];
+    float Fy[TN] = Fx * (Fdx - Fxdxsum) * scale;
+    TYPE y[TN] = Fy;
 
     // write-back
     *? (check)pdx = y;
@@ -145,36 +159,27 @@ class _sparse_softmax(torch.autograd.Function):
         offsets[1:] = torch.cumsum(sizes[:-1], dim=0)
         # block indices
         idx = torch.arange(layout.sum())
-        # rows
         rows = layout.nonzero()[:, 1]
-        # columns
         columns = layout.nonzero()[:, 2]
+        core   = torch.stack((idx, columns, rows), dim=1).view(-1)
         # construct look-up table
-        offsets += 2*sizes.numel()
+        offsets = offsets*3 + 2*sizes.numel()
         header = torch.stack((sizes, offsets), dim=1).view(-1)
-        lut = torch.cat((header, idx, columns, rows)).type(torch.int32).cuda()
-        return lut, sizes.max()
+        lut = torch.cat((header, core)).type(torch.int32).cuda()
+        return lut, int(sizes.max())
 
     @staticmethod
     def make_kernel(cache, src, max_k, dtype, block, kp_mask_mode, attn_mask_mode):
-        # pad tile to cover the entire reduction
-        params = {16384: (1, 32768, 16),
-                  8192:  (1, 16384, 16),
-                  4096:  (1, 8192, 16),
-                  2048:  (1, 4096, 16),
-                  1024:  (1, 2048, 16),
-                  512:   (1, 1024, 8),
-                  256:   (1, 512, 4),
-                  128:   (1, 256, 4)}
-        bound = max(128, 2**int(math.log2(max_k-1)))
-        if bound not in params:
+        if max_k >= 32768:
           raise NotImplementedError('Reductions larger than 32768 elements '\
                                     'are not yet implemented')
-        TM, TN, num_warps = params[bound]
+        num_warps = 4 if max_k < 512 else (4 if max_k < 2048 else 8)
+        pad = num_warps * 32 * 2
+        TN = (int(max_k) + pad-1)//pad * pad
         # just-in-time compile kernel
-        key = (dtype, TM, TN, num_warps, kp_mask_mode, attn_mask_mode)
+        key = (block, dtype, num_warps, TN, kp_mask_mode, attn_mask_mode)
         if key not in cache:
-            defines = {'TM': [TM], 'TN': [TN], 'TYPE': dtype, 'BLOCK': block,
+            defines = {'TM': [1], 'TN': [TN], 'TYPE': dtype, 'BLOCK': block,
                        'INFINITY': {torch.float32: 'F32_INFINITY',
                                     torch.float16: 'F16_INFINITY'}[dtype]}
             if kp_mask_mode == 'mul':
@@ -242,9 +247,9 @@ class Softmax:
                  key_padding_mask_mode='add', attn_mask_mode='add'):
         time_y = [None]
         if attn_mask is not None and attn_mask.dtype != x.dtype:
-            raise ValueError('Attention mask must be float32')
+            raise ValueError('Attention mask must be %s' % x.dtype)
         if key_padding_mask is not None and key_padding_mask.dtype != x.dtype:
-            raise ValueError('Key padding mask must be float32')
+            raise ValueError('Key padding mask must be %s' % x.dtype)
         x = Softmax.sparse_softmax(x, scale, key_padding_mask, attn_mask, 
                                   key_padding_mask_mode, attn_mask_mode,
                                   self.spdims, self.block,
